# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
from Game import GameState

def pieces_along_line(grid, pos, pnum, direction):
    """ Returns (number of pieces, steps to first enemy piece or 0 if none) """
    num_pieces = 0
    enemy_loc = 0
    dist = 0
    for p,val in GridUtils.enumerate_values_with_direction(grid, pos, direction):
        dist += 1
        if val>0:
            num_pieces += 1
            if enemy_loc==0 and val!=pnum:
                enemy_loc = dist
    return num_pieces, enemy_loc

class LinesOfActionGameState(GameState):

    def has_scores(self):
        return False

    def pieces_outside_largest_group(self, pnum):
        """ Returns the number of cells with value pnum that are not in the largest connected group """
        group_sizes = sorted(len(g) for g in GridUtils.connected_groups(self.grid, pnum, ortho=False))
        return sum(group_sizes[:-1])
        
    def is_game_over(self):
        return self.pieces_outside_largest_group(1)==0 or self.pieces_outside_largest_group(2)==0
        
    def winning_player(self):
        # if both players have all pieces connected, the last player who moved wins
        if self.pieces_outside_largest_group(self.previous_player())==0:
            return self.previous_player()
        if self.pieces_outside_largest_group(self.current_player)==0:
            return self.current_player
        return 0
    
    def successors(self):
        found_move = False
        pnum = self.current_player
        zeros = (0,)*self.grid.ndim
        for pos in self.enumerate_coords_with_value(pnum):
            for direction in GridUtils.direction_array(self.grid.ndim):
                if direction<zeros: continue
                # count pieces along line in both directions
                backdir = GridUtils.negate_tuple(direction)
                fwd_pieces, fwd_enemy = pieces_along_line(self.grid, pos, pnum, direction)
                bck_pieces, bck_enemy = pieces_along_line(self.grid, pos, pnum, backdir)
                total_pieces = fwd_pieces + bck_pieces + 1 # include self
                
                for movedir, enemy_dist in ((direction, fwd_enemy), (backdir, bck_enemy)):
                    # abort if blocked by enemy
                    if enemy_dist>0 and enemy_dist<total_pieces: continue
                    # good move if destination is valid and empty or enemy
                    incr = tuple(x*total_pieces for x in movedir)
                    dest = GridUtils.add_tuples(pos, incr)
                    if self.valid_coords(dest) and self.grid[dest]!=pnum:
                        found_move = True
                        nextgrid = self.grid.copy()
                        nextgrid[pos] = 0
                        nextgrid[dest] = pnum
                        yield ((pos,dest), self.next_state_with_grid(nextgrid))
        if not found_move:
            # this is rare, just pass
            yield ((), self.next_state_with_grid(nextgrid))

    def animations_for_move(self, move, newstate):
        # if nonempty move, always a move animation, capture if destination is nonempty
        import PGPieceAnimation
        if len(move)==2:
            animations = [PGPieceAnimation.move_animation(move[0], move[1])]
            if newstate.grid[move[1]]>0:
                animations.append(PGPieceAnimation.remove_animation(move[1]))
            return PGPieceAnimation.AnimationGroup(animations)
        return None


from GenericAI import GenericAI

class LinesOfActionAI (GenericAI):
    def game_utility(self, gstate):
        # piece outside largest group for opponent is good, self is bad
        return gstate.pieces_outside_largest_group(gstate.next_player()) - gstate.pieces_outside_largest_group(gstate.current_player)
                        
